/*--------------------------------------------------------------
	Image Processing functions    v2.1
		Written       by H.Goto , Jul 1994
		Last modified by H.Goto , Nov 1995
		Last modified by H.Goto , May 1996
		Last modified by H.Goto , Sep 1997
		Last modified by H.Goto , Oct 1998
		Last modified by H.Goto , Apr 2000
		Last modified by H.Goto , Nov 2001
		Last modified by H.Goto , Nov 2001
		Last modified by H.Goto , Feb 2002
		Last modified by H.Goto , May 2002
--------------------------------------------------------------*/

/*--------------------------------------------------------------------
  Copyright (C) 1994-2002  Hideaki Goto

        All Rights Reserved

  Permission to use, copy, modify, and distribute this software and
  its documentation for any purpose is hereby granted without fee,
  provided that (i) the above copyright notice and this permission
  notice appear in all copies and in supporting documentation, (ii)
  the name of the author, Hideaki Goto, may not be used in any
  advertising or otherwise to promote the sale, use or other
  dealings in this software without prior written authorization
  from the author, (iii) this software may not be used for
  commercial products without prior written permission from the
  author, and (iv) the notice of modification is specified in cases
  where modified copies of this software are distributed.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND.
  THE AUTHOR WILL NOT BE RESPONSIBLE FOR ANY DAMAGE CAUSED BY THIS
  SOFTWARE.
--------------------------------------------------------------------*/


#include	<stdio.h>
#include	<stdlib.h>
#include	<memory.h>

#define		siplib_c
#include	"siplib.h"


/*------------------------------------------------------
	Miscellaneous
------------------------------------------------------*/

				/* Get pointer to the specific scanline */

char *sip_getimgptr(SIPImage *image,ushort y){
	char	*p;
	p = image->data + (long)y * image->bytes_per_line;
	return(p);
}


				/* Get line size in bytes */

ushort sip_getlinebytes(ushort len,ushort depth,ushort pad){
	ushort	l;
	pad /= 8;
	if ( depth == 1 ){
		switch (pad){
		  case 2:	return( ((len +15) /16) *2 );
		  case 4:	return( ((len +31) /32) *4 );
		  default:	return( (len +7) /8 );
		}
	}
	l = len * (depth >> 3);
	switch (pad){
	  case 2:	return( ((l +1) /2) *2 );
	  case 4:	return( ((l +3) /4) *4 );
	  default:	return(l);
	}
}


				/* Copy ScanLine */

void sip_CopyLine(char *src,char *dst,ushort width,ushort depth){
	int	i;
	short	*sps,*dps;
	int32	*spl,*dpl;
	i = (int)width;
	if ( width < 8 ){
		switch (depth){
		  case 1:	i = (i + 7) /8;
		  case 8:
				for ( ; i>0 ; i-- )  *dst++ = *src++;
				break;
		  case 16:	sps = (short *)src;
				dps = (short *)dst;
				for ( ; i>0 ; i-- )  *dps++ = *sps++;
				break;
		  case 32:	spl = (int32 *)src;
				dpl = (int32 *)dst;
				for ( ; i>0 ; i-- )  *dpl++ = *spl++;
				break;
		}
		return;
	}
	switch (depth){
	  case 1:	i = (i + 7) /8;  break;
	  case 8:	break;
	  case 16:	i *= 2;  break;
	  case 32:	i *= 4;  break;
	}
	memcpy(dst,src,i);
	return;
}





/*------------------------------------------------------
	Image object manipulation
------------------------------------------------------*/


SIPImage	SIPImage0 = { 0,0,8,16,0,NULL };


				/* Create image */

SIPImage *sip_CreateImage(ushort width,ushort height,ushort depth){
	SIPImage	*imgtmp;
	long		linesize;
	long		bufsize;
	linesize = (long)((ulong)sip_getlinebytes(width,depth,16));
	bufsize = linesize * (long)((ulong)height);
	if ( bufsize  <= 0 )  return(NULL);
	if ( NULL == (imgtmp = (SIPImage *)malloc(sizeof(SIPImage))) ){
		return(NULL);
	}
	imgtmp->width  = width;
	imgtmp->height = height;
	imgtmp->depth  = depth;
	imgtmp->bitmap_pad  = 16;
	imgtmp->bytes_per_line  = linesize;
	if ( NULL == (imgtmp->data = (char *)malloc(bufsize)) ){
		free((char *)imgtmp);
		return(NULL);
	}
	return(imgtmp);
}


				/* Destroy image */

int sip_DestroyImage(SIPImage *image){
	if ( NULL != image ){
		if ( NULL != image->data ){
			free(image->data);
			image->data = NULL;
		}
		free((char *)image);
	}
	return(0);
}


				/* Fill a line by specified data (raw mode) */

int sip_rawClearLine(char *buf,ushort len,ushort depth,int32 data){
	int	c;
	char	db;
	char	*pc;
	short	*ps;
	int32	*pl;
	if ( data )  db = (char)0xff;  else db = 0;
	pl = (int32 *)( ps = (short *)( pc = buf ) );
	c = (int)len;
	switch ( depth ){
	  case 1:	c = (c +7) /8;
			for ( ; c>0 ; c-- )  *pc++ = db;
			break;
	  case 8:	for ( ; c>0 ; c-- )  *pc++ = (char)data;
			break;
	  case 16:	for ( ; c>0 ; c-- )  *ps++ = (short)data;
			break;
	  case 32:	for ( ; c>0 ; c-- )  *pl++ = (int32)data;
			break;
	}
	return(0);
}


				/* Fill a line by specified data */

int sip_ClearLine(SIPImage *image,ushort y,int32 data){
	char	*p;
	if ( NULL == image->data )  return(-1);
	if ( NULL == (p = sip_getimgptr(image,y)) )  return(-1);
	return( sip_rawClearLine(p,image->width,image->depth,data) );
}


				/* Fill image by specified data */

int sip_ClearImage(SIPImage *image,int32 data){
	int	c,y;
	char	db;
	char	*p,*pc;
	short	*ps;
	int32	*pl;
	if ( data )  db = (char)0xff;  else db = 0;
	if ( NULL == image->data )  return(-1);
	if ( NULL == (p = sip_getimgptr(image,0)) )  return(-1);
	for ( y=0 ; y < image->height ; y++ ){
		pl = (int32 *)( ps = (short *)(pc = p) );
		c = image->width;
		switch ( image->depth ){
		  case 1:	c = (c +7) /8;
				for ( ; c>0 ; c-- )  *pc++ = db;
				break;
		  case 8:	for ( ; c>0 ; c-- )  *pc++ = (char)data;
				break;
		  case 16:	for ( ; c>0 ; c-- )  *ps++ = (short)data;
				break;
		  case 32:	for ( ; c>0 ; c-- )  *pl++ = (int32)data;
				break;
		}
		p += image->bytes_per_line;
	}
	return(0);
}


				/* Copy area from src_image to dst_image */
				/*   1. depth,width,height must match    */
				/*   2. x,y,width,height must be         */
				/*         multiples of 8 for 1bit-image */

int sip_CopyArea(SIPImage *src,SIPRectangle *srect,
		 SIPImage *dst,SIPRectangle *drect){
	ushort	x,y;
	char	*sp,*dp;
	char	*spb,*dpb;
	short	*sps,*dps;
	int32	*spl,*dpl;
	ushort	depth;
	if ( src->depth    != dst->depth    )  return(-2);
	depth = src->depth;
	if ( srect->width  != drect->width  )  return(-2);
	if ( srect->height != drect->height )  return(-2);
	if ( (srect->x < 0) || (srect->y < 0) )  return(-1);
	if ( (drect->x < 0) || (drect->y < 0) )  return(-1);
	if ( (ushort)(srect->x + srect->width ) > src->width  )  return(0);
	if ( (ushort)(srect->y + srect->height) > src->height )  return(0);
	if ( (ushort)(drect->x + drect->width ) > dst->width  )  return(0);
	if ( (ushort)(drect->y + drect->height) > dst->height )  return(0);
	if ( NULL == (sp = sip_getimgptr(src,srect->y)) )  return(-1);
	if ( NULL == (dp = sip_getimgptr(dst,drect->y)) )  return(-1);
	if ( depth == 1 ){
		if ( srect->width & 0x07 )  return(-1);
		if ( srect->x & 0x07 )  return(-1);
		if ( drect->x & 0x07 )  return(-1);
		sp += srect->x >> 3;
		dp += drect->x >> 3;
	}
	else{	sp += (depth >> 3) * srect->x;
		dp += (depth >> 3) * drect->x;
	}
	for ( y=0 ; y < srect->height ; y++ ){
		spl = (int32 *)( sps = (short *)( spb = sp ) );
		dpl = (int32 *)( dps = (short *)( dpb = dp ) );
		switch (depth){
		  case 1:	for ( x=0 ; x < (srect->width >> 3) ; x++ )
								*dpb++ = *spb++;
				break;
		  case 8:	for ( x=0 ; x < srect->width ; x++ )  *dpb++ = *spb++;
				break;
		  case 16:	for ( x=0 ; x < srect->width ; x++ )  *dps++ = *sps++;
				break;
		  case 32:	for ( x=0 ; x < srect->width ; x++ )  *dpl++ = *spl++;
				break;
		}
		sp += src->bytes_per_line;
		dp += dst->bytes_per_line;
	}
	return(0);
}


				/* Get a scan line data */
				/*     len = -1  means "to end of line" */

int sip_GetScanLine(SIPImage *image, \
		void *buf,short x,short y,short len,ushort depth){
	ushort	i;
	char	*p;
	int	direct = 0;
	if ( (x < 0) || (x >= image->width ) )  return(-1);
	if ( (y < 0) || (y >= image->height) )  return(-1);
	if ( (len < 0) || (x + len) > image->width )  len = image->width - x;
	if ( NULL == (p = sip_getimgptr(image,y)) )  return(-1);
	if ( depth == image->depth )  direct = 1;
	if ( ! direct ){
		return(-2);
	}
	else{	switch (image->depth){
		  case 1:	if ( x   & 0x07 )  return(-1);
				p += x >> 3;
				if ( len & 0x07 )  return(-1);
				len >>= 3;
				for ( i=0 ; i<len ; i++ )
					*((char *)buf)++ = *p++;
				break;
		  case 8:	p += x;
				for ( i=0 ; i<len ; i++ )
					*((char *)buf)++ = *p++;
				break;
		  case 16:	p += 2 * x;
				for ( i=0 ; i<len ; i++ )
					*((short *)buf)++ = *((short *)p)++;
				break;
		  case 32:	p += 4 * x;
				for ( i=0 ; i<len ; i++ )
					*((int32 *)buf)++ = *((int32 *)p)++;
				break;
		}
	}
	return(0);
}


				/* Put a scan line data */
				/*     len = -1  means "to end of line" */

int sip_PutScanLine(SIPImage *image, \
		void *buf,short x,short y,short len,ushort depth){
	ushort	i;
	char	*p;
	int	direct = 0;
	if ( (x < 0) || (x >= image->width ) )  return(-1);
	if ( (y < 0) || (y >= image->height) )  return(-1);
	if ( (len < 0) || (x + len) > image->width )  len = image->width - x;
	if ( NULL == (p = sip_getimgptr(image,y)) )  return(-1);
	if ( depth == image->depth )  direct = 1;
	if ( ! direct ){
		return(-2);
	}
	else{	switch (image->depth){
		  case 1:	if ( x   & 0x07 )  return(-1);
				p += x >> 3;
				if ( len & 0x07 )  return(-1);
				len >>= 3;
				for ( i=0 ; i<len ; i++ )
					*p++ = *((char *)buf)++;
				break;
		  case 8:	p += x;
				memcpy(p,buf,len);
				break;
		  case 16:	p += 2 * x;
				for ( i=0 ; i<len ; i++ )
					*((short *)p)++ = *((short *)buf)++;
				break;
		  case 32:	p += 4 * x;
				for ( i=0 ; i<len ; i++ )
					*((int32 *)p)++ = *((int32 *)buf)++;
				break;
		}
	}
	return(0);
}


				/* Get a vertical scan line data */
				/*     len = -1  means "to end of line" */

int sip_GetVertLine(SIPImage *image, \
		void *buf,short x,short y,short len,ushort depth){
	ushort	i;
	char	*p;
	int	direct = 0;
	uchar	mask0,mask,d;
	if ( (x < 0) || (x >= image->width ) )  return(-1);
	if ( (y < 0) || (y >= image->height) )  return(-1);
	if ( (len < 0) || (y + len) > image->height )  len = image->height - y;
	if ( NULL == (p = sip_getimgptr(image,y)) )  return(-1);
	if ( depth == image->depth )  direct = 1;
	if ( ! direct ){
		return(-2);
	}
	else{	switch (image->depth){
		  case 1:	mask0 = (mask = (uchar)0x80) >> (x & 0x07);
				d = 0;
				if ( y   & 0x07 )  return(-1);
				p += (x >> 3);
				for ( i=0 ; i<len ; i++ ){
					if ( mask0 & *p )  d |= mask;
					p += image->bytes_per_line;
					if ( 0 == (mask >>= 1) ){
						*(++(char *)buf) = d;
						mask = (uchar)0x80;
					}
				}
				if ( mask != (uchar)0x80 )  *((char *)buf)++ = d;
				break;
		  case 8:	p += x;
				for ( i=0 ; i<len ; i++ ){
					*((char *)buf)++ = *p;
					p += image->bytes_per_line;
				}
				break;
		  case 16:	p += 2 * x;
				for ( i=0 ; i<len ; i++ ){
					*((short *)buf)++ = *((short *)p);
					p += image->bytes_per_line;
				}
				break;
		  case 32:	p += 4 * x;
				for ( i=0 ; i<len ; i++ ){
					*((int32 *)buf)++ = *((int32 *)p);
					p += image->bytes_per_line;
				}
				break;
		}
	}
	return(0);
}


				/* Put a vertical scan line data */
				/*     len = -1  means "to end of line" */

int sip_PutVertLine(SIPImage *image, \
		void *buf,short x,short y,short len,ushort depth){
	ushort	i;
	char	*p;
	int	direct = 0;
	uchar	mask0,mask,d0,d;
	if ( (x < 0) || (x >= image->width ) )  return(-1);
	if ( (y < 0) || (y >= image->height) )  return(-1);
	if ( (len < 0) || (y + len) > image->height )  len = image->height - y;
	if ( NULL == (p = sip_getimgptr(image,y)) )  return(-1);
	if ( depth == image->depth )  direct = 1;
	if ( ! direct ){
		return(-2);
	}
	else{	switch (image->depth){
		  case 1:	mask0 = (mask = (uchar)0x80) >> (x & 0x07);
				d = *((char *)buf);
				if ( y   & 0x07 )  return(-1);
				p += (x >> 3);
				for ( i=0 ; i<len ; i++ ){
					d0 = *p & ~mask0;
					if ( mask & d )  d0 |= mask0;
					*p = d0;
					p += image->bytes_per_line;
					if ( 0 == (mask >>= 1) ){
						d = *(++(char *)buf);
						mask = (uchar)0x80;
					}
				}
				break;
		  case 8:	p += x;
				for ( i=0 ; i<len ; i++ ){
					*p = *((char *)buf)++;
					p += image->bytes_per_line;
				}
				break;
		  case 16:	p += 2 * x;
				for ( i=0 ; i<len ; i++ ){
					*((short *)p) = *((short *)buf)++;
					p += image->bytes_per_line;
				}
				break;
		  case 32:	p += 4 * x;
				for ( i=0 ; i<len ; i++ ){
					*((int32 *)p) = *((int32 *)buf)++;
					p += image->bytes_per_line;
				}
				break;
		}
	}
	return(0);
}


				/* Get pixel value on bitmap image */
uchar sip_getbit(SIPImage *image,ushort x,ushort y){
	uchar	b;
	uchar	mask;
	mask = 0x01 << (7 - (x & 0x07));
	b = *(image->data + (int32)(x>>3) + (int32)y * image->bytes_per_line) & mask;
	return( (uchar)(b != 0) );
}


				/* Put pixel value on bitmap image */
uchar sip_putbit(SIPImage *image,ushort x,ushort y,uchar b){
	uchar	*p;
	uchar	mask;
	mask = 0x01 << (7 - (x & 0x07));
	if ( b != 0 )  b = (uchar)0xff;
	p = (uchar *)image->data + (int32)(x>>3) + (int32)y * image->bytes_per_line;
	*p = (*p & ~mask) | (b & mask);
	return(b);
}


				/* OR & put pixel value on bitmap image */
uchar sip_orbit(SIPImage *image,ushort x,ushort y,uchar b){
	uchar	*p;
	uchar	mask;
	mask = 0x01 << (7 - (x & 0x07));
	if ( b != 0 )  b = (uchar)0xff;
	p = (uchar *)image->data + (int32)(x>>3) + (int32)y * image->bytes_per_line;
	*p = *p | (b & mask);
	return(b);
}


				/* AND & put pixel value on bitmap image */
uchar sip_andbit(SIPImage *image,ushort x,ushort y,uchar b){
	uchar	*p;
	uchar	mask;
	mask = 0x01 << (7 - (x & 0x07));
	if ( b != 0 )  b = (uchar)0xff;
	p = (uchar *)image->data + (int32)(x>>3) + (int32)y * image->bytes_per_line;
	*p = *p & (b & mask);
	return(b);
}


				/* XOR & put pixel value on bitmap image */
uchar sip_xorbit(SIPImage *image,ushort x,ushort y,uchar b){
	uchar	*p;
	uchar	mask;
	mask = 0x01 << (7 - (x & 0x07));
	if ( b != 0 )  b = (uchar)0xff;
	p = (uchar *)image->data + (int32)(x>>3) + (int32)y * image->bytes_per_line;
	*p = *p ^ (b & mask);
	return(b);
}


 

/*------------------------------------------------------
	Data conversion
------------------------------------------------------*/

				/* Depth conversion from 1bit to 8bit */

void sip_cvt1to8(char *src,ushort bitoff,char *dst,ushort len,char fc,char bc){
	ushort	i;
	uchar	d;
	uchar	mask;
	if ( bitoff ){
		mask = 0x80 >> bitoff;
		bitoff = 8 - bitoff;
		d = (uchar)*src++;
		for ( i=0 ; (i < bitoff) && (len > 0) ; i++, len--, mask >>= 1 ){
			if ( d & mask )  *dst++ = fc;  else *dst++ = bc;
		}
	}
	for ( i = len >> 3 ; i>0 ; i-- ){
		d = (uchar)*src++;
		if ( d & 0x80 )  *dst++ = fc;  else *dst++ = bc;
		if ( d & 0x40 )  *dst++ = fc;  else *dst++ = bc;
		if ( d & 0x20 )  *dst++ = fc;  else *dst++ = bc;
		if ( d & 0x10 )  *dst++ = fc;  else *dst++ = bc;
		if ( d & 0x08 )  *dst++ = fc;  else *dst++ = bc;
		if ( d & 0x04 )  *dst++ = fc;  else *dst++ = bc;
		if ( d & 0x02 )  *dst++ = fc;  else *dst++ = bc;
		if ( d & 0x01 )  *dst++ = fc;  else *dst++ = bc;
	}
	mask = (uchar)0x80;
	d = (uchar)*src++;
	for ( i = len & 0x07 ; i>0 ; i-- ){
		if ( d & mask )  *dst++ = fc;  else *dst++ = bc;
		mask >>= 1;
	}
}


				/* Depth conversion from 1bit to 16bit */

void sip_cvt1to16(char *src,ushort bitoff,short *dst,ushort len,short fc,short bc){
	ushort	i;
	uchar	d;
	uchar	mask;
	if ( bitoff ){
		mask = 0x80 >> bitoff;
		bitoff = 8 - bitoff;
		d = (uchar)*src++;
		for ( i=0 ; (i < bitoff) && (len > 0) ; i++, len--, mask >>= 1 ){
			if ( d & mask )  *dst++ = fc;  else *dst++ = bc;
		}
	}
	for ( i = len >> 3 ; i>0 ; i-- ){
		d = (uchar)*src++;
		if ( d & 0x80 )  *dst++ = fc;  else *dst++ = bc;
		if ( d & 0x40 )  *dst++ = fc;  else *dst++ = bc;
		if ( d & 0x20 )  *dst++ = fc;  else *dst++ = bc;
		if ( d & 0x10 )  *dst++ = fc;  else *dst++ = bc;
		if ( d & 0x08 )  *dst++ = fc;  else *dst++ = bc;
		if ( d & 0x04 )  *dst++ = fc;  else *dst++ = bc;
		if ( d & 0x02 )  *dst++ = fc;  else *dst++ = bc;
		if ( d & 0x01 )  *dst++ = fc;  else *dst++ = bc;
	}
	mask = (uchar)0x80;
	d = (uchar)*src++;
	for ( i = len & 0x07 ; i>0 ; i-- ){
		if ( d & mask )  *dst++ = fc;  else *dst++ = bc;
		mask >>= 1;
	}
}


				/* Depth conversion from 1bit to 32bit */

void sip_cvt1to32(char *src,ushort bitoff,int32 *dst,ushort len,int32 fc,int32 bc){
	ushort	i;
	uchar	d;
	uchar	mask;
	if ( bitoff ){
		mask = 0x80 >> bitoff;
		bitoff = 8 - bitoff;
		d = (uchar)*src++;
		for ( i=0 ; (i < bitoff) && (len > 0) ; i++, len--, mask >>= 1 ){
			if ( d & mask )  *dst++ = fc;  else *dst++ = bc;
		}
	}
	for ( i = len >> 3 ; i>0 ; i-- ){
		d = (uchar)*src++;
		if ( d & 0x80 )  *dst++ = fc;  else *dst++ = bc;
		if ( d & 0x40 )  *dst++ = fc;  else *dst++ = bc;
		if ( d & 0x20 )  *dst++ = fc;  else *dst++ = bc;
		if ( d & 0x10 )  *dst++ = fc;  else *dst++ = bc;
		if ( d & 0x08 )  *dst++ = fc;  else *dst++ = bc;
		if ( d & 0x04 )  *dst++ = fc;  else *dst++ = bc;
		if ( d & 0x02 )  *dst++ = fc;  else *dst++ = bc;
		if ( d & 0x01 )  *dst++ = fc;  else *dst++ = bc;
	}
	mask = (uchar)0x80;
	d = (uchar)*src++;
	for ( i = len & 0x07 ; i>0 ; i-- ){
		if ( d & mask )  *dst++ = fc;  else *dst++ = bc;
		mask >>= 1;
	}
}


				/* Depth conversion from 8bit to 16bit */
				/*                        (unsigned)   */

void sip_cvt8to16u(uchar *src,ushort *dst,ushort len,ushort mul){
	ushort	i;
	if ( mul == 1 ){
		for ( i=0 ; i<len ; i++ )  *dst++ = (ushort)*src++;
	}
	else{	for ( i=0 ; i<len ; i++ )  *dst++ = (ushort)*src++ * mul;
	}
}


				/* Depth conversion from 8bit to 16bit */
				/*                         (signed)    */

void sip_cvt8to16(char *src,short *dst,ushort len,short mul){
	ushort	i;
	if ( mul == 1 ){
		for ( i=0 ; i<len ; i++ )  *dst++ = (short)*src++;
	}
	else{	for ( i=0 ; i<len ; i++ )  *dst++ = (short)*src++ * mul;
	}
}


				/* Depth conversion from 8bit to 32bit */
				/*                        (unsigned)   */

void sip_cvt8to32u(uchar *src,uint32 *dst,ushort len,uint32 mul){
	ushort	i;
	if ( mul == 1 ){
		for ( i=0 ; i<len ; i++ )  *dst++ = (uint32)*src++;
	}
	else{	for ( i=0 ; i<len ; i++ )  *dst++ = (uint32)*src++ * mul;
	}
}


				/* Depth conversion from 8bit to 32bit */
				/*                         (signed)    */

void sip_cvt8to32(char *src,int32 *dst,ushort len,int32 mul){
	ushort	i;
	if ( mul == 1 ){
		for ( i=0 ; i<len ; i++ )  *dst++ = (int32)*src++;
	}
	else{	for ( i=0 ; i<len ; i++ )  *dst++ = (int32)*src++ * mul;
	}
}


				/* Depth conversion from 16bit to 32bit */
				/*                         (unsigned)   */

void sip_cvt16to32u(ushort *src,uint32 *dst,ushort len,uint32 mul){
	ushort	i;
	if ( mul == 1 ){
		for ( i=0 ; i<len ; i++ )  *dst++ = (uint32)*src++;
	}
	else{	for ( i=0 ; i<len ; i++ )  *dst++ = (uint32)*src++ * mul;
	}
}


				/* Depth conversion from 16bit to 32bit */
				/*                          (signed)    */

void sip_cvt16to32(short *src,int32 *dst,ushort len,int32 mul){
	ushort	i;
	if ( mul == 1 ){
		for ( i=0 ; i<len ; i++ )  *dst++ = (int32)*src++;
	}
	else{	for ( i=0 ; i<len ; i++ )  *dst++ = (int32)*src++ * mul;
	}
}


				/* Depth conversion from 8bit to 1bit */
				/*                          (signed)  */

void sip_cvt8to1(char *src,char *dst,ushort bitoff,ushort len,char th){
	ushort	i,len8;
	char	d;
	uchar	m;
	if ( bitoff != 0 ){
		m = (uchar)((int)0x80 >> bitoff);
		d = *dst;
		for ( i = 8 - bitoff ; i > 0 && len > 0 ; i--, len-- ){
			if ( *src > th )  d |= m;  else d &= ~m;
			++src;
			m >>= 1;
		}
		*dst++ = d;
	}
	len8 = len / 8;
	len = len - len8 * 8;
	for ( ; len8 > 0 ; len8-- ){
		d = 0;
		if ( *src > th )  d |= (char)0x80;
		++src;
		if ( *src > th )  d |= (char)0x40;
		++src;
		if ( *src > th )  d |= (char)0x20;
		++src;
		if ( *src > th )  d |= (char)0x10;
		++src;
		if ( *src > th )  d |= (char)0x08;
		++src;
		if ( *src > th )  d |= (char)0x04;
		++src;
		if ( *src > th )  d |= (char)0x02;
		++src;
		if ( *src > th )  d |= (char)0x01;
		++src;
		*dst++ = d;
	}
	if ( len != 0 ){
		d = *dst;
		m = 0x80;
		for ( ; len > 0 ; len-- ){
			if ( *src > th )  d |= m;  else d &= ~m;
			++src;
			m >>= 1;
		}
		*dst++ = d;
	}
}


				/* Depth conversion from 8bit to 1bit */
				/*                         (unsigned) */

void sip_cvt8to1u(uchar *src,uchar *dst,ushort bitoff,ushort len,uchar th){
	ushort	i,len8;
	uchar	d;
	uchar	m;
	if ( bitoff != 0 ){
		m = 0x80 >> bitoff;
		d = *dst;
		for ( i = 8 - bitoff ; i > 0 && len > 0 ; i--, len-- ){
			if ( *src > th )  d |= m;  else d &= ~m;
			++src;
			m >>= 1;
		}
		*dst++ = d;
	}
	len8 = len / 8;
	len = len - len8 * 8;
	for ( ; len8 > 0 ; len8-- ){
		d = 0;
		if ( *src > th )  d |= (uchar)0x80;
		++src;
		if ( *src > th )  d |= (uchar)0x40;
		++src;
		if ( *src > th )  d |= (uchar)0x20;
		++src;
		if ( *src > th )  d |= (uchar)0x10;
		++src;
		if ( *src > th )  d |= (uchar)0x08;
		++src;
		if ( *src > th )  d |= (uchar)0x04;
		++src;
		if ( *src > th )  d |= (uchar)0x02;
		++src;
		if ( *src > th )  d |= (uchar)0x01;
		++src;
		*dst++ = d;
	}
	if ( len != 0 ){
		d = *dst;
		m = 0x80;
		for ( ; len > 0 ; len-- ){
			if ( *src > th )  d |= m;  else d &= ~m;
			++src;
			m >>= 1;
		}
		*dst++ = d;
	}
}


				/* Depth conversion from 16bit to 1bit */
				/*                          (signed)    */

void sip_cvt16to1(short *src,char *dst,ushort bitoff,ushort len,short th){
	ushort	i,len8;
	char	d;
	char	m;
	if ( bitoff != 0 ){
		m = 0x80 >> bitoff;
		d = *dst;
		for ( i = 8 - bitoff ; i > 0 && len > 0 ; i--, len-- ){
			if ( *src > th )  d |= m;  else d &= ~m;
			++src;
			m >>= 1;
		}
		*dst++ = d;
	}
	len8 = len / 8;
	len = len - len8 * 8;
	for ( ; len8 > 0 ; len8-- ){
		d = 0;
		if ( *src > th )  d |= (char)0x80;
		++src;
		if ( *src > th )  d |= (char)0x40;
		++src;
		if ( *src > th )  d |= (char)0x20;
		++src;
		if ( *src > th )  d |= (char)0x10;
		++src;
		if ( *src > th )  d |= (char)0x08;
		++src;
		if ( *src > th )  d |= (char)0x04;
		++src;
		if ( *src > th )  d |= (char)0x02;
		++src;
		if ( *src > th )  d |= (char)0x01;
		++src;
		*dst++ = d;
	}
	if ( len != 0 ){
		d = *dst;
		m = (char)0x80;
		for ( ; len > 0 ; len-- ){
			if ( *src > th )  d |= m;  else d &= ~m;
			++src;
			m >>= 1;
		}
		*dst++ = d;
	}
}


				/* Depth conversion from 16bit to 8bit */
				/*                        (unsigned)   */

void sip_cvt16to8u(ushort *src,uchar *dst,ushort len,ushort mul,ushort div){
	ushort	i;
	if ( mul == 1 ){
		if ( div == 1 )
			for ( i=0 ; i<len ; i++ )  *dst++ = (uchar)*src++;
		else	for ( i=0 ; i<len ; i++ )  *dst++ = (uchar)(*src++ / div);
	}
	else{	if ( div == 1 )
			for ( i=0 ; i<len ; i++ )  *dst++ = (uchar)(*src++ * mul);
		else	for ( i=0 ; i<len ; i++ )  *dst++ = (uchar)((*src++ * mul) / div);
	}
}




/*------------------------------------------------------
	Convolution
------------------------------------------------------*/

static int sip_convolution3x3s(SIPMASKs *mask,IPStream *ips,short *buf){
	int	i;
	short	d;
	short	m00,m10,m20;
	short	m01,m11,m21;
	short	m02,m12,m22;
	short	*p0,*p1,*p2;
	m00 = mask->mask[0];	m10 = mask->mask[1];	m20 = mask->mask[2];
	m01 = mask->mask[3];	m11 = mask->mask[4];	m21 = mask->mask[5];
	m02 = mask->mask[6];	m12 = mask->mask[7];	m22 = mask->mask[8];
	p0 = (short *)ips->line[0];
	p1 = (short *)ips->line[1];
	p2 = (short *)ips->line[2];
	for ( i = (int)ips->width ; i > 0 ; i-- ){
		d  = *p0 * m00 + *(p0+1) * m10 + *(p0+2) * m20;
		d += *p1 * m01 + *(p1+1) * m11 + *(p1+2) * m21;
		d += *p2 * m02 + *(p2+1) * m12 + *(p2+2) * m22;
		*buf++ = d / mask->div;
		++p0;  ++p1;  ++p2;
	}
	return(0);
}


int sip_convolution_s(SIPMASKs *mask,IPStream *ips,short *buf){
	int	i,j,x;
	int	masksize,width,ms1,ms2,ms4;
	short	d;
	short	*m,*p;
	masksize = (int)mask->masksize;
/*	if ( mask->depth != 16 )  return(-1);
*/	if ( masksize == 3 ){
		return ( sip_convolution3x3s(mask,ips,buf) );
	}

	ms4 = masksize /4;
	ms2 = masksize & 0x02;
	ms1 = masksize & 0x01;
	width = (int)ips->width;
	for ( x=0 ; x < width ; x++ ){
		d = 0;
		for ( j=0 ; j<masksize ; j++ ){
			m = mask->mask + (j * masksize);
			p = (short *)ips->line[j] + x;
			for ( i=0 ; i<ms4 ; i++ ){
				d += *p * *m + *(p+1) * *(m+1) + *(p+2) * *(m+2) + *(p+3) * *(m+3);
				p += 4;  m += 4;
			}
			if ( ms2 ){
				d += *p * *m + *(p+1) * *(m+1);
				p += 2;  m += 2;
			}
			if ( ms1 )  d += *p * *m;
		}
		*buf++ = d / mask->div;
	}
	return(0);
}


int sip_convolution(SIPMASK *mask,IPStream *ips,char *buf){
	switch ( mask->s.depth ){
	  case 16:	return( sip_convolution_s(&mask->s,ips,(short *)buf) );
/*	  case 32:	return( sip_convolution_l(&mask->l,ips,(int32 *)buf) );
	  case 0x8020:	return( sip_convolution_f(&mask->f,ips,(float *)buf) );
*/	}
	return(-1);
}




/*------------------------------------------------------
	Clipping
------------------------------------------------------*/

void clip8(char *src,char *dst,ushort len,char lo,char hi){
	char	d;
	for ( ; len > 0 ; len-- ){
		d = *src++;
		if ( d < lo )  d = lo;
		if ( d > hi )  d = hi;
		*dst++ = d;
	}
}


void clip8u(uchar *src,uchar *dst,ushort len,uchar lo,uchar hi){
	uchar	d;
	for ( ; len > 0 ; len-- ){
		d = *src++;
		if ( d < lo )  d = lo;
		if ( d > hi )  d = hi;
		*dst++ = d;
	}
}


void clip16(short *src,short *dst,ushort len,short lo,short hi){
	short	d;
	for ( ; len > 0 ; len-- ){
		d = *src++;
		if ( d < lo )  d = lo;
		if ( d > hi )  d = hi;
		*dst++ = d;
	}
}


void clip16u(ushort *src,ushort *dst,ushort len,ushort lo,ushort hi){
	ushort	d;
	for ( ; len > 0 ; len-- ){
		d = *src++;
		if ( d < lo )  d = lo;
		if ( d > hi )  d = hi;
		*dst++ = d;
	}
}


void clipint(int *src,int *dst,ushort len,int lo,int hi){
	int	d;
	for ( ; len > 0 ; len-- ){
		d = *src++;
		if ( d < lo )  d = lo;
		if ( d > hi )  d = hi;
		*dst++ = d;
	}
}


void clipuint(uint *src,uint *dst,ushort len,uint lo,uint hi){
	uint	d;
	for ( ; len > 0 ; len-- ){
		d = *src++;
		if ( d < lo )  d = lo;
		if ( d > hi )  d = hi;
		*dst++ = d;
	}
}




